/*
 * bookmark.h
 * vim: expandtab:ts=4:sts=4:sw=4
 *
 * Copyright (C) 2012 - 2019 James Booth <boothj5@gmail.com>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later WITH OpenSSL-exception
 */

#ifndef XMPP_BOOKMARK_H
#define XMPP_BOOKMARK_H

#include <glib.h>

void bookmark_request(void);

#endif
