/*
 * test_message.h
 *
 * Copyright (C) 2015 James Booth <boothj5@gmail.com>
 * Copyright (C) 2026 - 2026 Michael Vetter <jubalh@iodoru.org>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */
void message_send(void** state);
void message_receive_console(void** state);
void message_receive_chatwin(void** state);
