/*
 * test_roster.h
 *
 * Copyright (C) 2015 James Booth <boothj5@gmail.com>
 * Copyright (C) 2026 - 2026 Michael Vetter <jubalh@iodoru.org>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */
void sends_new_item(void** state);
void sends_new_item_nick(void** state);
void sends_remove_item(void** state);
void sends_remove_item_nick(void** state);
void sends_nick_change(void** state);
