/*
 * test_i18n.h
 *
 * Copyright (C) 2026 Michael Vetter <jubalh@iodoru.org>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */
void i18n_msg_nickname(void** state);
void i18n_win_nickname(void** state);
void i18n_autocomplete_tab_utf8(void** state);
void i18n_autocomplete_tab_latin(void** state);
