/*
 * screen.h
 * vim: expandtab:ts=4:sts=4:sw=4
 *
 * Copyright (C) 2012 - 2019 James Booth <boothj5@gmail.com>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later WITH OpenSSL-exception
 */

int screen_titlebar_row(void);
int screen_statusbar_row(void);
int screen_inputwin_row(void);

int screen_mainwin_row_start(void);
int screen_mainwin_row_end(void);
