/*
 * test_presence.h
 *
 * Copyright (C) 2015 - 2017 James Booth <boothj5@gmail.com>
 * Copyright (C) 2026 - 2026 Michael Vetter <jubalh@iodoru.org>
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */
void presence_away(void** state);
void presence_away_with_message(void** state);
void presence_online(void** state);
void presence_online_with_message(void** state);
void presence_xa(void** state);
void presence_xa_with_message(void** state);
void presence_dnd(void** state);
void presence_dnd_with_message(void** state);
void presence_chat(void** state);
void presence_chat_with_message(void** state);
void presence_set_priority(void** state);
void presence_includes_priority(void** state);
void presence_keeps_status(void** state);
void presence_received(void** state);
void presence_missing_resource_defaults(void** state);
